/*
 * GWT - General Windowing Toolkit
 *
 * Copyright (C) 1997 Uwe Maurer - uwe_maurer@t-online.de
 *                    Andreas Beck - becka@ggi-project.org
 *               1998 Rodolphe Ortalo - ortalo@laas.fr
 *                    Tristan Wibberley - bloater@ps.cus.umist.ac.uk
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: init.c,v $
 * Revision 1.11  2000/06/09 11:34:44  ortalo
 * Added support for drawing functions based on LibXMI.
 *
 * Revision 1.10  2000/03/19 19:32:46  ortalo
 * Some debugging and modifications (wrt OpenAmulet mainly).
 *
 * Revision 1.9  2000/03/13 22:02:55  ortalo
 * Committed new version of LibGWT corresponding to 0.0.2.
 * The event management API is probably more or less defined. The simplest
 * approach finally won. (I should have remembered: KISS!)
 *
 * Revision 1.8  1999/05/24 00:48:37  marcus
 * Fixed broken APPASSERT statement.
 *
 * Revision 1.7  1999/05/24 00:30:56  marcus
 * Use LibGG instead of pthread mutexes.
 *
 * Revision 1.6  1999/05/13 20:01:38  ortalo
 * New version of libgwt. Now using LibGII for input management.
 * Some drawing functions implemented. Visibility management should
 * be rather okay. Still some work to do, but now, your eyes can
 * look at some drawings... not only code. ;-)
 *
 * Revision 1.5  1998/11/17 23:28:08  ortalo
 * Brought libgwt under autoconf&automake&libtool, in conformance with
 * the GNU configure and building system (and also, the rest of GGI. :-)
 * Should build and run as before (i.e.: not so much... :-(.)
 *
 * Revision 1.4  1998/11/17 15:13:28  tristan
 *
 * Changed my email address
 *
 * Revision 1.3  1998/08/15 16:15:30  marcus
 * Fixed to compile with -pedantic
 *
 * Revision 1.2  1998/08/11 22:10:08  ortalo
 * Incorporated the changes made by Tristan Wibberley. The code is
 * still in alpha development state.
 *
 * Revision 1.1  1998/07/09 18:38:39  ortalo
 * Initial integration of libgwt in the GGI repository.
 * This is experimental code, for your eyes only !
 * (Don't execute... ;-)
 *
 */

#include "internal.h"
#include "gwt_debug.h"
#include <ggi/gg.h>
#include <ggi/gii.h>
#include <ggi/ggi.h>
#include <ggi/xmi.h>
#include <ggi/ggi_ext.h>
#include <string.h>

/*
 *
 *   Initialisation segment of libgwt
 *
 */

/*
 * Internal structures of initialisation of library
 */

static int	_gwtIsUp = 0; /* starter flag */
ggi_extid	gwtID = -1;   /* ext. ID (default to -1 should segfault easily) */
void      	*_gwt_global_lock = NULL; /* Global mutex */

/* Debug flags: everything disabled by default */
uint32 _gwtDebugState = 0;
int _gwtDebugSync = 0;

/*********************************************************************
 *                  INTERNAL FUNCTIONS AND PROCEDURES                *
 *********************************************************************/

/*
 * Prints one windows main characteristics on STDERR
 */
static void gwt_print_one_window(gwt_window_t w)
{
  ggi_coord tl, br;
  gwt_region_bounds(w->geometry, &tl, &br);
  fprintf(stderr,"win(%p):left=%i,top=%i,width=%i,height=%i (hidden=%i)",
	  w, tl.x, tl.y, (br.x - tl.x), (br.y - tl.y), w->hidden);
  return;
}
/*
 * Prints a window hierarchy on STDERRE - handles indentation
 */
static void gwt_print_windows(gwt_window_t w, unsigned int indentlevel)
{
  gwt_window_t child;
  unsigned int il;
  for (il = indentlevel; il > 0; il--) fprintf(stderr,"   ");
  fprintf(stderr,"-> ");
  gwt_print_one_window(w);
  fprintf(stderr,"\n");
  /* Print childs */
  child = GWT_Z_TOP_CHILD(w);
  while (child != NULL)
    {
      gwt_print_windows(child, indentlevel + 1);
      child = GWT_Z_DOWN(child);
    }
  return;
}

/*
 * Change notification mechanism
 */
static int changed(ggi_visual_t vis, int whatchanged)
{
  GWTDPRINT(" changed called - vis=%p, what=%i\n",
	 vis,whatchanged);

  switch (whatchanged) {
  case GGI_CHG_APILIST:
    {
      int temp;
      char api[256], args[256];
      for (temp=0; 0==ggiGetAPI(vis,temp,api,args); temp++) {
	strcat(api, "-gwt");
	GWTDPRINT(" would now load #%d: %s(%s)\n",temp,api,args);
      }
    }
  break;
  }
  return 0;
}

/*********************************************************************
 *                  EXPORTED FUNCTIONS AND PROCEDURES                *
 *********************************************************************/

/************************
 * Initialisation & co. *
 ************************/

/*
 * Initialisation function
 *  REQUIRES: _gwtIsUp be initialised to 0.
 */
int gwtInit(void)
{
  int err;
  const char* str;

  if (_gwtIsUp > 0) return 0; /* Init. only on first call */

  if ((_gwt_global_lock = ggLockCreate()) == NULL) {
	  fprintf(stderr,"LibGWT: unable to initialize global mutex.\n");
	  return GGI_EUNKNOWN;
  }

  /* Initialize the GII library. */
  err = ggiInit();
  if (err) {
    GWTDPRINT_CORE("gwtInit: unable to initialize LibGII.\n");
    ggLockDestroy(_gwt_global_lock);
    return err;
  }

  /* Initialize the GGI library. */
  err = ggiInit();
  if (err) {
    GWTDPRINT_CORE("gwtInit: unable to initialize LibGGI.\n");
    giiExit();
    ggLockDestroy(_gwt_global_lock);
    return err;
  }

  /* Initialize the XMI library. */
  err = xmiInit();
  if (err) {
    GWTDPRINT_CORE("gwtInit: unable to initialize LibGGI.\n");
    ggiExit();
    giiExit();
    ggLockDestroy(_gwt_global_lock);
    return err;
  }

  gwtID = ggiExtensionRegister("GWT", sizeof(gwtprivate), changed);
  if (gwtID < 0) {
    GWTDPRINT_CORE("gwtInit: unable to register extension (err=%d)\n", gwtID);
    ggiExit();
    giiExit();
    ggLockDestroy(_gwt_global_lock);
    return gwtID;
  }

  GWTDPRINT(" Real initialisation\n");
  _gwtIsUp++;

  str = getenv("GWT_DEBUG");
  if (str != NULL) {
    _gwtDebugState = atoi(str);
    GWTDPRINT_CORE("Debugging(GWT)=%d\n", _gwtDebugState);
  }
  str = getenv("GWT_DEBUGSYNC");
  if (str != NULL) {
    _gwtDebugSync = 1;
  }
		
  return GGI_OK; /* Successful termination */
}

/*
 * Exit function
 */
int gwtExit(void)
{
  int ret;

  LIBGWT_APPASSERT((_gwtIsUp > 0),
		   "gwtExit called but nothing initialised yet!");

  _gwtIsUp--;
  if (_gwtIsUp > 0) return 0; /* Exit really only on last call */

  GWTDPRINT(" Final real exit\n");

  ret = ggiExtensionUnregister(gwtID);
  ggiExit();
  giiExit();
  ggLockDestroy(_gwt_global_lock);

  return ret;
}
/*
 * Attaches LibGWT as a GGI visual extension
 */
int gwtAttach(ggi_visual_t vis)
{
	int ret;
	GWTDPRINT(" Preattached to %p.\n", vis);
	ret=ggiExtensionAttach(vis, gwtID);
	GWTDPRINT(" Attached to %p. ret=%i\n", vis, ret);

	if (ret==0) { /* We are actually creating the primary instance */
	              /* assign initial LIBGGI_EXT(vis, gwtID) private data */
	  /* TODO: This seems to cause a segfault under the debugger ? -- ortalo
		strcpy(LIBGGI_EXT(vis, gwtID), "GWT Private Data !");
	  */
		/* Now fake an "API change" so the right libs get loaded */
		changed(vis, GGI_CHG_APILIST);
	}

	return ret;
}
/*
 * Detach LibGWT from the visual
 */
int gwtDetach(ggi_visual_t vis)
{
	int ret;
	ret=ggiExtensionDetach(vis, gwtID);
	GWTDPRINT(" Detached from %p. ret=%i\n", vis, ret);

	return ret;
}

/************************************
 * Debugging and printing functions *
 ************************************/
/*
 * Prints a window hierarchy on stderr
 */
int gwtDebugPrintWindows(gwt_window_t win)
{
  if (win != NULL)
    gwt_print_windows(win, 0);
  else
    fprintf(stderr,"-> win(0x0000000) !!!\n");
  return 0;
}
